print("import: 'thriftpy'")
import thriftpy

print("import: 'thriftpy.contrib'")
import thriftpy.contrib

print("import: 'thriftpy.contrib.tracking'")
import thriftpy.contrib.tracking

print("import: 'thriftpy.parser'")
import thriftpy.parser

print("import: 'thriftpy.protocol'")
import thriftpy.protocol

print("import: 'thriftpy.transport'")
import thriftpy.transport

print("import: 'thriftpy.transport.buffered'")
import thriftpy.transport.buffered

print("import: 'thriftpy.transport.framed'")
import thriftpy.transport.framed

print("import: 'thriftpy.transport.memory'")
import thriftpy.transport.memory

print("import: 'thriftpy.protocol.cybin'")
import thriftpy.protocol.cybin

